'use strict';

//the name of KEY should be uppercase
const DEFAULT_OPERATOR_MAPPING = {
    'CONTAINS': {
        operator: 'like', template: function (value) {
            return '%' + value + '%';
        }
    },
    'STARTSWITH': {
        operator: 'like', template: function (value) {
            return value + '%';
        }
    },
    'ENDSWITH': {
        operator: 'like', template: function (value) {
            return '%' + value;
        }
    },
    'LIKE': {
        operator: 'like', template: function (value) {
            return '%' + value + '%';
        }
    },
    'AFTER': '>',
    'BEFORE': '<',
    'GT': '>',
    'LT': '<',
    'GTE': '>=',
    'LTE': '<=',
    'EQ': '=',
    'NOTCONTAINS': {
        operator: 'not like', template: function (value) {
            return '%' + value + '%';
        }
    },
    'NOTSTARTSWITH': {
        operator: 'not like', template: function (value) {
            return value + '%';
        }
    },
    'NOTENDSWITH': {
        operator: 'not like', template: function (value) {
            return '%' + value;
        }
    },
    'NOTLIKE': {
        operator: 'not like', template: function (value) {
            return '%' + value + '%';
        }
    },
    'NOT': '<>',
    'ISNULL': {
        operator: 'is null', whereType:'WhereNull', template: function (value) {
            return HttpQueryParser.IS_NULL_VALUE;
        }
    },
    'ISNOTNULL': {
        operator: 'is not null', whereType:'WhereNotNull', template: function (value) {
            return HttpQueryParser.IS_NOT_NULL_VALUE;
        }
    },
    'ISEMPTY': {
        operator: '=', template: function (value) {
            return '';
        }
    },
    'ISNOTEMPTY': {
        operator: '<>', template: function (value) {
            return '';
        }
    },
    'IN': {
        operator: 'in', whereType:'WhereIn',template: function (value) {
            if(typeof value === 'string'){
                value = value.split(',')
            }
            return value;
        }
    },
    'NOTIN': {
        operator: 'not in', whereType:'WhereNotIn', template: function (value) {
            if(typeof value === 'string'){
                value = value.split(',')
            }
            return value;
        }
    },
    'BETWEEN': {
        operator: 'between', whereType:'WhereBetween',template: function (value) {
            if(typeof value === 'string'){
                value = value.split(',')
            }
            return value.map(e=>e*1);
        }
    },
    'NOTBETWEEN': {
        operator: 'not between', whereType:'WhereNotBetween', template: function (value) {
            if(typeof value === 'string'){
                value = value.split(',')
            }
            return value.map(e=>e*1);
        }
    },

};

const OPERATOR_MAP = new Map();

for (let key in DEFAULT_OPERATOR_MAPPING) {
    OPERATOR_MAP.set(key, DEFAULT_OPERATOR_MAPPING[key]);
}

/**
 * 转换http参数到sql参数,查询API的核心逻辑
 */
class HttpQueryParser {

    static IS_NULL_VALUE = Symbol();
    static IS_NOT_NULL_VALUE = Symbol();

    constructor(options) {
        options = options || {};
        this.skipEmptyValue = options.hasOwnProperty('skipEmptyValue') ? options.skipEmptyValue : true;
    }

    /**
     * http queryString 转换器
     * @param httpQueryStringObj
     * @returns {{}}
     */
    parse(httpQueryStringObj) {

        let parsedQueryObj = {};

        if (httpQueryStringObj) {
            for (let key in httpQueryStringObj) {

                let value = httpQueryStringObj[key];

                //if it is null or empty
                if (!this.isQueryForNull(key) && typeof value === 'string' && (!value || !value.trim())) {
                    this.treatEmptyParameter(key, value, httpQueryStringObj);
                } else {
                    parsedQueryObj[key] = this.translateOperator(key, value)
                }

            }
        }
        return parsedQueryObj;

    }

    isQueryForNull(key){
        let keyAndOperator = key.split(':');
        return keyAndOperator.length > 1 && ['ISNULL','ISNOTNULL','ISEMPTY','ISNOTEMPTY'].some(e=>e == keyAndOperator[1].toUpperCase());
    }

    /**
     * 空参数处理策略
     * @param key
     * @param value
     * @param parsedResult
     */
    treatEmptyParameter(key, value, parsedResult) {
        //TODO 应实现为一个策略对象如SKIP_EMPTY_VALUE_STRATEGY,but not now
        if (this.skipEmptyValue) {
            delete parsedResult[key];
        }
    }

    /**
     * 转换查询操作符号
     * @param key
     * @param value
     * @returns {*[]}
     */
    translateOperator(key, value) {

        let keyAndOperator = key.split(':');
        let whereType = 'Where';

        if (keyAndOperator.length > 1) {

            let knexPattern = this.adaptKeyToKnexPattern(keyAndOperator);

            if (OPERATOR_MAP.has(knexPattern.operatorAlias)) {

                let operatorMapping = OPERATOR_MAP.get(knexPattern.operatorAlias);
                var operatorAlias = operatorMapping;

                if (typeof operatorMapping == 'object') {
                    operatorAlias = operatorMapping.operator;
                    if(!operatorMapping.whereType){
                        operatorMapping.whereType = 'Where';
                    }

                    let template = operatorMapping.template;

                    if (template && typeof  template == 'function') {
                        value = template.call(template, value);
                        whereType = operatorMapping.whereType;
                    }
                }
            }

            return {
                whereType:  whereType,
                columnName: knexPattern.columnName,
                operatorAlias: operatorAlias,
                value: value,
                condition : knexPattern.condition
                };

        } else {
            return {
                whereType : whereType,
                columnName: key,
                value: value,
                operatorAlias: '=',
                condition : 'and'
            };
        }
    }

    /**
     *
     * 将http query的 key适配成knex可识别的模式
     * @param key query参数的key
     */
    adaptKeyToKnexPattern(key) {

        let isOrCondition = key[0].toUpperCase() === 'OR';

        let columnName = key[isOrCondition ? 1 : 0];//[or, columnName, (operator)]
        let operatorAlias = isOrCondition ? (key.length === 3 ? key[2] : 'eq') : key[1];

        return {
            columnName: columnName,
            operatorAlias: operatorAlias.toUpperCase(),
            condition: isOrCondition ? 'or' : 'and'
        };
    }


}


module.exports = HttpQueryParser;


