"use strict";

const debug = require('debug')('datahub-core:knex-helper');
const rp = require('request-promise');
const co = require('co');
const Promise = require('bluebird');
const fsExtra = Promise.promisifyAll(require('fs-extra'));
const fs = Promise.promisifyAll(require('fs'));
const events = require('events');

// constants
const META_TABLE_NAME = 'db_connection';
const METAKEY_TABLE_NAME = 'db_key';
const SQL_ACCESS_LOG_TABLE_NAME = 'sql_access_log';
const CDB_PREFIX = 'cdb_';
const META_CONFIG_DB_KEY = '_meta';
const CACHE_KEY_DB_TREE = '/tmp/datahubTree.js';

const driver = 'mysql';

let instance = null;

/**
 * KNEX的实用方法集合
 */
class KnexHelper extends events.EventEmitter {

    constructor(opts) {

        super();

        this.opts = opts || {};

        this.databaseConnectionCache = new Map();
        this.databaseConfigCache = {};
        this.metaDataCahce = new Map();
        this.freeDatabase = new Map();

        this.metaName = opts.name || META_CONFIG_DB_KEY;
        this.metaDatabaseName = opts.metaDatabaseName || 'datahub';
        this.metaTable = `${this.metaDatabaseName}.${META_TABLE_NAME}`;
        this.metaKeyTable = `${this.metaDatabaseName}.${METAKEY_TABLE_NAME}`;

        this.accessLogTable = `${this.metaDatabaseName}.${SQL_ACCESS_LOG_TABLE_NAME}`;
        this.apiProperty = opts.apiProperty;
        this.env = opts.env;

        this.insecureAuth = opts.insecureAuth === undefined ? true : opts.insecureAuth;

        this.knex = require('knex')(this.opts.options);
        this.knex.on('query-error', function (error, obj) {
            debug(error, obj);
        });

        this.pollingInterval = this.opts.pollingInterval || 60000;

        this.initMetaTable()
            .then(() => {
                this.metaDataCahce.clear();
                return this.cacheMetaData();
            }).then(() => {
            return this.loadMetaData();
        }).catch((err) => {
            debug(err);
        });

        //定时刷新数据库链接
        //由于涉及js对象和链接状态,所以不做序列化到远程缓存,而是存到进程内缓存
        process.nextTick(this.polling());
        setInterval(this.polling(), this.pollingInterval);

        //数据库树结构生成成本高,缓存起来
        //FIXME slow!!
        //process.nextTick(this.createCallback(this, this.buildDatabasesTree));
        //setInterval(this.createCallback(this, this.buildDatabasesTree), 3600 * 1000);

        this.on('reloadCache', (data) => {
            console.log('reloadCache executing');
            this.polling();
        });
    }

    //singleton
    static getInstance(options) {
        if (!instance || !(instance instanceof KnexHelper)) {
            return new KnexHelper(options);
        } else {
            return instance;
        }
    }

    /* ===meta数据的缓存 ===*/
    loadMetaData() {
        let ctx = this;

        for (var databaseConfig of this.metaDataCahce.values()) {
            const cdbKey = databaseConfig.cdb_key
            const needUpdate = databaseConfig.needUpdate
            if (this.databaseConnectionCache.has(cdbKey) && needUpdate) {
                debug('connection changed', databaseConfig);
                try {
                    this.databaseConnectionCache.get(cdbKey).destroy()
                } catch (error) {
                    debug(error);
                }
                this.databaseConnectionCache.delete(cdbKey)
            }

            if (!this.databaseConnectionCache.has(databaseConfig.cdb_key)) {
                debug('Creating connection', databaseConfig);
                try {
                    this.databaseConnectionCache.set(databaseConfig.cdb_key, this._connect(databaseConfig, {min: 0, max: 5}));
                    this.databaseConfigCache[databaseConfig.cdb_key] = databaseConfig;
                } catch (e) {
                    debug(e);
                }
            }
        }

        this.opts.options.connection.cdb_key = this.metaName;
        this.metaDataCahce.set(this.metaName, this.opts.options.connection);
        debug('MetaDataCache is %o', this.opts);

        for (let cdbKey of this.databaseConnectionCache.keys()) {
            if(!this.metaDataCahce.has(cdbKey)) {
                debug('connection delete', cdbKey);
                try {
                    this.databaseConnectionCache.get(cdbKey).destroy()
                } catch (error) {
                    debug(error);
                }
                this.databaseConnectionCache.delete(cdbKey)
            }
        }

    }

    addDatabaseConnection(config, key, chartset) {
        let databases = this.getMetaDataCache();
        let nConf = Object.assign({}, config);

        if (!databases.has(key)) {
            try {
                nConf.cdb_key = key;
                nConf.names = chartset;
                this.databaseConnectionCache.set(key, this._connect(nConf, {min: 0, max: 5}));
                this.metaDataCahce.set(key, nConf);
                this.databaseConfigCache[key] = nConf;
            } catch (e) {
                debug(e);
            }
        }

        return this.databaseConnectionCache.get(key);
    }

    getMetaDataCache() {
        return this.metaDataCahce;
    }

    getDatabaseConnections() {
        return this.databaseConnectionCache;
    }

    getDatabaseConnection(key) {
        return this.databaseConnectionCache.get(key);
    }

    getDatabaseConfig(key) {
        return this.databaseConfigCache[key];
    }

    cacheMetaData() {

        this.databaseConnectionCache.set(this.metaName, this.knex);

        return this.getMetaFromDB().then((resp) => {
            resp.map(e => {
                if (e.conn_status) {
                    if(this.metaDataCahce.has(e.cdb_key)) {
                        const old = this.metaDataCahce.get(e.cdb_key)
                        if(old.host !== e.host || old.port !== e.port || old.user !== e.user || old.password !== e.password) {
                            e.needUpdate = true
                        }
                    }else {
                        e.needUpdate = true
                    }
                    this.metaDataCahce.set(e.cdb_key, e);
                    if(!e.owner) {
                        this.freeDatabase.set(e.cdb_key, true);
                    }else {
                        this.freeDatabase.delete(e.cdb_key);
                    }
                } else {
                    this.metaDataCahce.delete(e.cdb_key);
                    this.freeDatabase.delete(e.cdb_key);
                }

            });
            debug('Cached meta table:', this.metaDataCahce);
        }).catch((e) => {
            debug(e);
        });
    }

    getDatabaseConnectionConfigs() {
        this.databaseConnectionCache.set(this.metaName, this.knex);
        return this.getMetaFromDB();
    }

    getMetaConnection() {
        return this.databaseConnectionCache.get(this.metaName);
    }

    polling() {
        return () => {
            debug('Polling meta data.');
            this.cacheMetaData()
                .then(() => {
                    return this.loadMetaData();
                }).catch((err) => {
                debug(err);
            });
        };
    }

    /* ===meta数据的构建 ===*/

    /**
     * 初始化meta数据
     */
    initMetaTable() {

        var promise = new Promise((resolve, reject) => {
            this.knex.schema.hasTable(this.metaTable).then((exists) => {
                debug('Is table exists? %s', exists);
                if (!exists) {
                    reject('create table ' + this.metaTable + ' first');
                } else {
                    resolve(exists);
                }
            }, (err) => {
                reject(err);
            });
        });
        return promise;
    }

    _connect(opt, pool) {

        let poolDef = {
            afterCreate: function (connection, callback) {
                //fix http://git.code.oa.com/Capture/webapp_datahub/issues/23
                connection.query("SET time_zone = '+8:00';", function (err) {
                    callback(err, connection);
                });
            }
        };

        Object.assign(poolDef, pool || {});

        let databaseConnection = require('knex')({
            client: driver,
            connection: {
                host: opt.host,
                user: opt.user,
                password: opt.password,
                port: opt.port,
                insecureAuth: this.insecureAuth,
                supportBigNumbers: true,
                bigNumberStrings: true
            },
            acquireConnectionTimeout: 5000,
            pool: poolDef,
            log: {
                warn(message) {
                    debug('knex warn', message);
                },
                error(message) {
                    debug('knex error', message);
                },
                deprecate(message) {
                    debug('knex deprecate', message);
                },
            }
        });

        databaseConnection.on('query-error', function (error, obj) {
            debug(error, obj);
        });

        return databaseConnection;
    }


    * saveDatabaseConfig(entity) {
        return yield this.knex(this.metaTable).insert(entity);
    }

    getMetaFromDB(includeInvalid = true) {
        return new Promise((resolve, reject) => {
            let q = this.knex.select('id', 'host', 'port', 'user', 'password', 'conn_status', 'owner')
                .from(this.metaTable);
            if(!includeInvalid){
                q.where({"conn_status": 1});
            }
                q.then((resp) => {
                    resp.forEach(e => e['cdb_key'] = CDB_PREFIX + e.id);
                    resolve(resp);
                }).catch(reject);
        });
    }

    /**
     * 构建数据库树状结构
     * @returns {IterableIterator<*>}
     */
    * buildDatabasesTree() {
        return []

        let databases = yield this.getDatabaseConnectionConfigs(false);
        let url = this.getInternalApiUrl();

        debug("Building databases tree", databases);

        for (let db of databases) {

            db['schemas'] = [];

            let cdbKey = db.cdb_key;
            debug("Building databases tree node", db, url);

            try {

                const headers = {
                    "x-datahub-host": "datahub.wsd.oa.com",
                    "x-tracer-sampler-rate": 0
                };


                let schemas = yield rp({uri: url + '/' + cdbKey, json: true, headers: headers});

                for (let schema of schemas) {

                    schema['tables'] = [];
                    delete schema['href'];

                    debug("Building databases tree leaf node", schema.schema);

                    try {
                        let tables = yield rp({
                            uri: url + '/' + cdbKey + '/' + schema.schema, json: true, headers: headers
                        });

                        tables.forEach((e) => delete e['href']);
                        schema['tables'] = tables;

                    } catch (e) {
                        debug('Failed to fetch tables of', cdbKey, schema.schema);
                    }
                }
                db['schemas'] = schemas;

            } catch (e) {
                debug('Failed to fetch schemas of', cdbKey, e);
            }
        }

        try {
            yield fsExtra.writeJsonAsync(CACHE_KEY_DB_TREE, databases);
        } catch (e) {
            debug(e);
        }

        debug('Save to cache', databases);

        return databases;
    }

    /**
     * 获取树状状态，带缓存
     * @returns {IterableIterator<*>}
     */
    * getDatabasesTree() {

        let tree = [];

        try {
            let stat = yield fs.statAsync(CACHE_KEY_DB_TREE);
            debug(stat.mtime);
            if (stat.size === 0 || new Date - stat.mtime > 3600 * 1000) {
                debug(`File change time ${stat.mtime} is expired or file is empty`);
                tree = yield this.buildDatabasesTree();
            } else {
                tree = yield fsExtra.readJsonAsync(CACHE_KEY_DB_TREE);
            }

        } catch (e) {
            debug(e);
            tree = yield this.buildDatabasesTree();
        }

        let dbKeys = yield this.getMetaConnection()
            .select(this.metaKeyTable + '.id as db_key_id', this.metaKeyTable + '.*', this.metaTable + '.*')
            .from(this.metaKeyTable).leftJoin(this.metaTable, this.metaTable + '.id', this.metaKeyTable + '.connection_id');

        let treeForDbKey = [];

        dbKeys.forEach((e) => {
            tree.forEach((t) => {
                if ('cdb_' + e.connection_id == t.cdb_key) {
                    let ele = JSON.parse(JSON.stringify(t));
                    ele.cdb_key = e.db_key;
                    treeForDbKey.push(ele);
                }
            });
        });

        return tree.concat(treeForDbKey);

    }


    /* ===sql服务 ===*/
    /**
     * 链接测试器
     * @param host
     * @param port
     * @param user
     * @param password
     * @returns {IterableIterator<*>}
     */
    * tryConnect({host, port, user, password}) {

        let testConnection = require('knex')({
            client: driver,
            connection: {
                host: host,
                user: user,
                password: password,
                port: port,
                insecureAuth: this.insecureAuth,
                supportBigNumbers: true,
                bigNumberStrings: true,
                multipleStatements: true
            },
        })

        try {
            console.log(testConnection);
            testConnection.client.pool.check();
            let res = yield testConnection.raw('select 1');
            return {res: res, connect: true};
        } catch (err) {
            debug('test ', err);
            return {connect: false, err: err};
        } finally {
            testConnection.destroy();
        }
    }

    /**
     * 只读执行器
     * @returns {Function}
     */
    static readOnlyTx() {

        /**
         * 事务promise化
         */
        let promiseTransaction = function (scope) {

            return new Promise(function (fulfill, reject) {
                scope.state.connection.transaction(function (trx) {
                    try {
                        fulfill(trx)
                    } catch (err) {
                        reject(err);
                    } finally {
                        return trx;
                    }
                });
            });
        }

        return function* (next) {
            let trx = yield promiseTransaction(this);
            this.state.trx = trx;
            try {
                yield next;
            } catch (err) {
                throw err;
            } finally {
                trx.rollback();
            }
            debug('Http get method not support transaction, rollback now.');
        };
    }

    /**
     * 自定义sql的分页支持
     */
    * paginate(sql, query, connection) {

        let limit = query.pageSize * 1 || query.count * 1 || 30;
        limit = limit > 1000 ? 1000 : limit;
        let page = query.page * 1 || 0;
        page = page < 0 ? 0 : page;

        let offset = page * limit;
        delete query['pageSize'];
        delete query['page'];
        delete query['_lego_schema'];

        query.limit = limit;
        query.offset = offset;

        let param = Object.assign({}, query);
        let result = yield connection.raw(
            sql + ' limit :limit offset :offset', param);
        let count = yield connection.raw('select count(*) as c from (' + sql + ') as d', param);

        return {result: result[0], options: {limit: limit, count: count[0][0].c, page: page}};

    };


    /* === API辅助操作，依赖getMetaConnection ===*/

    /**
     * 主键
     * @param connection
     * @param schema
     * @param table
     * @returns {IterableIterator<*>}
     */
    * showPrimaryKeys(connection, schema, table) {

        let indexes = yield connection.raw('SHOW INDEX FROM :table:', {table: schema + '.' + table});
        let primaryKeys = indexes[0].filter(e => e.Key_name == 'PRIMARY');

        if (primaryKeys.length > 0) {
            return primaryKeys.map(e => e.Column_name);
        } else {
            return [];
        }
    }

    /**
     * 获取链接
     * @param dbKey
     * @returns {IterableIterator<*>}
     */
    * findConnectionByDbkey(dbKey) {
        let result = yield this.getMetaConnection()
            .select('connection_id').from(this.metaKeyTable)
            .where({db_key: dbKey});

        if (result.length > 0) {
            return CDB_PREFIX + result[0].connection_id;
        } else {
            return 0;
        }
    }

    /**
     * 查询别名
     * @param keyword
     * @param limit
     * @param page
     * @returns {IterableIterator<*>}
     */
    * findAlias(keyword, limit, page) {

        page = page < 0 ? 0 : page;
        let offset = page * limit;

        let query = this.getMetaConnection()
            .select(this.metaKeyTable + '.id as db_key_id', this.metaKeyTable + '.*', this.metaTable + '.*')
            .from(this.metaKeyTable).leftJoin(this.metaTable, this.metaTable + '.id', this.metaKeyTable + '.connection_id');
        let sql = (keyword ?
            query.where(this.metaTable + '.host', 'like', '%' + keyword + '%')
                .orWhere(this.metaTable + '.port', 'like', '%' + keyword + '%')
                .orWhere(this.metaKeyTable + '.db_key', 'like', '%' + keyword + '%')
            : query);

        let aliases = yield sql.limit(limit).offset(offset);

        let count = yield this.getMetaConnection()
            .count('* as c')
            .from(this.metaKeyTable).leftJoin(this.metaTable, this.metaTable + '.id', this.metaKeyTable + '.connection_id');

        return {result: aliases, options: {limit: limit, count: count[0].c, page: page}};

    }

    /* === meta数据操作，依赖getMetaConnection ===*/

    /**
     * log down sql access information
     * @param ip
     * @param accessTime
     * @param sqlMappingId
     * @param sql
     * @param query
     * @param source
     * @returns {*}
     */
    * logSqlAccessing(ip, accessTime, sqlMappingId, sql, query, source) {
        let metaConn = this.getMetaConnection();
        let log = {
            ip: ip,
            sql_mapping_id: sqlMappingId,
            sql: sql,
            query: query,
            access_time: accessTime,
            cost_time: -1, //not execute yet
            source: source
        };

        return yield metaConn(this.accessLogTable).insert(log);
    }

    /**
     * update sql cost time of access log
     * @param id
     * @param costTimeInMs
     * @returns {*}
     */
    * updateCostTime(id, costTimeInMs) {
        let metaConn = this.getMetaConnection();
        return yield metaConn(this.accessLogTable).where({id: id})
            .update({cost_time: costTimeInMs});
    }

    /* ===参数转换拼接===*/

    createCallback(ctx, gen, cb) {
        function error(err) {
            if (err) throw err
        }

        return function () {
            cb = cb || error;
            co.call(ctx, gen).then(function () {
                cb()
            }, cb)
        }
    }

    getInternalApiUrl() {

        let port = this.apiProperty.port;
        let host = this.apiProperty.host;
        let endpoint = this.apiProperty.endpoint;

        return `http://${host}:${port}${endpoint}`;
    }

    getMetaUrl() {
        return `${this.apiProperty.endpoint}/${this.metaName}/${this.opts.options.connection.database}`;
    }

    getOptions() {
        return this.opts.options;
    }

}

module.exports = KnexHelper;
