"use strict";

const debug = require('debug')('datahub-core:mount');

const knexHelper = require('./core/knex-helper');

const entityRoute = require('./routes/entities-router');
const schemaRoute = require('./routes/schemas-router');
const restfulRoute = require('./routes/restfuls-router');
const DatabaseRoute = require('./routes/databases-router');
const sqlMappingRoute = require('./routes/sql-mappings-router');
const sqlExecutorRoute = require('./routes/sql-executors-router');

const OpLogMiddleware = require('./middlewares/oplog-middleware');
const hookExecutor = require('./middlewares/hook-executor');
const privilege = require('./middlewares/privilege')


class DatahubRestfulCore {

    constructor(config) {

        config = config || {};

        this.endpoint = config.endpoint || '/api/v1';
        this.env = config.env || 'dev';
        this.options = config.options || {}
        this.genericSession = config.genericSession || require('koa-generic-session')();
        this.port = config.port || 3000;
        this.host = config.host || 'localhost';
        this.name = config.name || 'datahub';
        this.bodyLimit = config.bodyLimit || '4mb'
        this.app = config.app || require('koa')();

        debug('options:', this.options);
        debug('port:', this.port);
        debug('host:', this.host);
        debug('endpoint:', this.endpoint);
        debug('env:', this.env);
        debug('name:', this.name);

        let kcfg = {
            apiProperty: {port: this.port, host: this.host, endpoint: this.endpoint},
            name: this.name,
            options: this.options
        };
        this.knexHelperInstance = knexHelper.getInstance(kcfg);
        debug('core init:', kcfg);
    }

    mount() {

        let endpoint = this.endpoint;
        let knexHelperInstance = this.knexHelperInstance;
        this.app.context.knexHelper = knexHelperInstance

        this.app.use(privilege);
        this.app.use(new OpLogMiddleware({knexHelperInstance}).middleware());
        this.app.use(require('./core/http/export-support').middleware());
        this.app.use(require('./routes/appinfo-router').routes());

        //钩子执行器
        const executor = new hookExecutor({endpoint, knexHelperInstance});
        this.app.use(executor.middleware());

        //逻辑执行器
        this.app.use(require('./routes/vm-router')({endpoint, knexHelperInstance}).routes());

        //chaining route, providing /modules/schema/entity/.* style url
        entityRoute.use('/:entity', restfulRoute.routes(), restfulRoute.allowedMethods());
        schemaRoute.use('/:schema/_q', sqlMappingRoute.routes(), sqlMappingRoute.allowedMethods());
        schemaRoute.use('/:schema', entityRoute.routes(), entityRoute.allowedMethods());

        let databaseRoute = DatabaseRoute({endpoint: this.endpoint});

        databaseRoute.use('/:database/sqlExecutor', sqlExecutorRoute.routes(), sqlExecutorRoute.allowedMethods());
        databaseRoute.use('/:database'
            , schemaRoute.routes()
            , schemaRoute.allowedMethods());

        this.app.use(databaseRoute.routes());

        this.app.start = (cb) => {
            return this.app.listen(this.port, this.host, () => {
                console.log('Datahub core started http://%s:%s%s%s', this.host, this.port, this.endpoint, this.name);
                cb && cb();
            });
        };

        return this.app;
    }
}

module.exports = {DatahubRestfulCore, knexHelper};
