'use strict';

const debug = require('debug')('datahub-core:hook:debug');
const vm = require('vm');

const HOOK_KEY = '_hook';
const HOOK_AFTER = ':after';
const HOOK_BEFORE = ':before';
const HOOK_AROUND = ':around';

class HookExecutor {

    constructor(options) {
        debug('options', options);
        this.options = options;
    }

    get endpoint(){
        return this.options.endpoint;
    }

    /**
     * 解析hook and 清理query string的hook参数
     * @param hookString
     */
    resolveHook(query) {

        let hookObj = {};

        this.appendHookString(query, HOOK_KEY + HOOK_AFTER, hookObj);
        this.appendHookString(query, HOOK_KEY + HOOK_BEFORE, hookObj);
        this.appendHookString(query, HOOK_KEY + HOOK_AROUND, hookObj);

        //Clear hook for query
        delete query[HOOK_KEY + HOOK_AFTER];
        delete query[HOOK_KEY + HOOK_AROUND];
        delete query[HOOK_KEY + HOOK_BEFORE];

        debug('Hook obj is %o', hookObj);

        return hookObj;
    }

    /**
     * 根据hook key追加hook串
     * @param query
     * @param key
     * @param hookObj
     */
    appendHookString(query, key, hookObj) {
        if (query[key]) {
            let hookKey = key.replace(HOOK_KEY + ":", "");
            hookObj[hookKey] = query[key].split('|').map((e) => {
                return {scriptName: e};
            });
        }
    }

    /**
     * 是否有hook
     * @param query
     * @returns {boolean}
     */
    hasHook(query) {
        return !!query[HOOK_KEY + HOOK_AFTER] || !!query[HOOK_KEY + HOOK_BEFORE] || !!query[HOOK_KEY + HOOK_AROUND];
    }

    /**
     * datahub暴露的中间件
     * @returns {* Function}
     */
    middleware() {

        let path = this.options.endpoint;
        let self = this;

        return function *(next) {

            let koaContext = this;

            if (koaContext.request.path.search(path) > -1) {

                let hasHook = self.hasHook(koaContext.query);

                if (hasHook) {

                    let hookObj = self.resolveHook(koaContext.query);

                    //share sandbox between hooks
                    let sandBox = vm.createContext({
                        "require": require,
                        "console": console,
                        module: {exports: {}}
                    });

                    if (hookObj.before) {
                        yield self.executeScript(hookObj.before, sandBox, koaContext, next);
                    }

                    //环绕的hook必须实现yield next
                    if (hookObj.around) {
                        yield self.executeScript(hookObj.around, sandBox, koaContext, next);
                    } else {
                        yield next;
                    }

                    if (hookObj.after) {
                        yield self.executeScript(hookObj.after, sandBox, koaContext, next);
                    }
                }

            }

            yield next;
        }
    }

    /**
     * 查出hook脚本内容
     * @param hookScripts
     * @returns {*}
     */
    * getCodeSnippet(hookScripts) {

        let conn = this.options.knexHelperInstance.getMetaConnection();
        let scriptNames = hookScripts.map((e) => e.scriptName);

        return yield conn.select('code_snippet', 'hook_name').from('hook_config').where(
            'hook_name', 'in', scriptNames
        );
    }

    /**
     * 注入作用域并执行hook
     * @param hookScripts
     * @param sandBox
     * @param koaContext
     * @param next
     */
    * executeScript(hookScripts, sandBox, koaContext, next) {
        let script = yield this.getCodeSnippet(hookScripts);
        if (script) {
            try {
                let generator = this.runInVM(script, next, sandBox, koaContext);
                yield generator;
            } catch (e) {
                koaContext.status = 500;
                koaContext.type = 'html';
                koaContext.body = `<h1>${e.message}</h1><code>${e.stack}</code>`;
                debug("Failed to execute hook %o", e);
            }
        }
    }

    /**
     * script执行的VM实现
     * @param scripts
     * @param next
     * @param context
     * @param koaCtx
     *
     * @return generator
     */
    * runInVM(scripts, next, context, koaCtx) {

        for (let script of scripts) {

            let hookCodeSnippet = script.code_snippet;

            let wrapFunctionText = `
                (function(){
                   return function *(next){
                    ${hookCodeSnippet}
                   }
                })
             `;

            let scriptName = `hook/virtual/${script.hook_name}.js`;

            debug('Evaluating script %s', scriptName);

            let wrapFunction = vm.runInContext(wrapFunctionText, context, {filename: scriptName})();
            yield wrapFunction.bind(koaCtx)(next);

        }

    }

}

module.exports = HookExecutor;