'use strict';

const BasicAuthorizationProvider = require('../core/http/basic-authorization-provider');
const authProvider = new BasicAuthorizationProvider();
const debug = require('debug')('datahub-core:oplog:debug');
const uuid = require('uuid').v4;

const setDefaultUser = function(koaCtx){

    if(!koaCtx.session){
        koaCtx.session = {};
    }
    if (koaCtx.session && !koaCtx.session.user) {
        koaCtx.session.user = {LoginName: koaCtx.state.httpAuth.username};
    }


}

/**
 * Operation log to db koa middleware
 */
class OplogMiddleware {

    constructor(options) {
        this.knexHelperInstance = options.knexHelperInstance;
    }

    * save(log) {
        let metaConn = this.knexHelperInstance.getMetaConnection();
        yield metaConn('opr_log').insert(log);
    }

    middleware() {

        let self = this;

        return function* (next) {

            let startTime = new Date().getTime();

            let authInfo = authProvider.adaptToKoa(this).authorize();
            this.state.httpAuth = authInfo;

            setDefaultUser(this);

            try {
                this.opLog = function* (startTime, operation, operationDesc, extMsg, status) {
                    let log = self.internalLogInfo.bind(this)(startTime, operation, operationDesc, extMsg, status || 200);
                    yield self.save(log);
                }
            } catch (e) {
                //noop
            }

            let error = 0;

            try {
                yield next;
            } catch (e) {
                error = e;
                throw e;

            } finally {

                let sampler = this.headers['x-tracer-sampler-rate'] || 1.00;

                if ((this.method == 'POST' ||
                    this.method == 'PUT' ||
                    this.method == 'PATCH' ||
                    this.method == 'DELETE') && sampler > 0) {

                    let log = self.internalLogInfo.bind(this)(startTime);
                    if (error) {

                        this.status = error.status || 500;
                        log.result = error.status || 500;
                        log.fal_rpt = error.status || 500;
                        log.op_desc = error.message;
                        log.ext_msg = JSON.stringify(error.stack);
                    }
                    debug('Logging operation', log);
                    yield self.save(log);
                }
            }

        }
    }

    internalLogInfo(opTime, operation, operationDesc, extMsg, status) {

        extMsg = typeof extMsg !== 'string' ? JSON.stringify(extMsg) : extMsg;

        return {
            request_id: this.state._requestId || uuid(),
            username: this.session.user.LoginName,
            client_ip: this.ip,
            operation: (operation || (this.path + this.querystring)).substring(0, 127),
            op_desc: operationDesc || '',
            log_category: this.app.name || '',
            result: status || this.status,
            fal_rpt: (status || this.status) < 400 ? 0 : this.status,
            delta_time: new Date().getTime() - opTime,
            op_time: opTime,
            ext_msg: extMsg || JSON.stringify({method: this.method, host: this.host, body: this.request.body})
        };
    }

}

module.exports = OplogMiddleware;