'use strict';

const debug = require('debug')('krh:route');
const route = require('koa-router');

const httpQueryParser = require('../core/http/http-query-parser');
const util = require('./util');
const datetimeService = require('../core/utils/datetime');
const co = require('co');
const jsonFilter = require('../core/utils/json-key-filter');

const importSupport = require('../core/http/import-support');
/**
 * Very basic curd notificationRouter
 * @type {Router|exports|module.exports}
 */

var router = new route();
let queryLogger = function (data) {
    let ctx = this;
    return co(function* () {
        yield ctx.opLog(new Date().getTime(), 'sql:' + data.method, data.sql, data.bindings, 100);
    });
};

let updateFunction = function* () {

    let body = yield datetimeService.transformDatetimeFields(
        this.state.connection,
        this.state.schema,
        this.state.entity,
        this.request.body);

    yield this.state.connection(this.state.entity)
        .withSchema(this.state.schema)
        .where(this.state.pkParams)
        .update(body)
        .on('query', queryLogger.bind(this));

    let entity = yield this.state.connection(this.state.entity)
        .withSchema(this.state.schema)
        .where(this.state.pkParams).limit(1);

    this.body = entity[0];

};

router
    .get('/meta/pk', function* () {
        //Primary keys detection
        this.body = yield this.knexHelper.showPrimaryKeys(this.state.connection, this.state.schema, this.state.entity);

    })
    .get('/meta/columns(.*)', function* () {
        let body = yield this.state.connection(this.state.entity)
            .select('*').from('information_schema.columns')
            .where({'table_name': this.state.entity, 'table_schema': this.state.schema});
        this.body = body;
    })
    .all('(.*)', function* (next) {

        yield next;

    })
    .all('/:id', function* (next) {

        let idField = yield this.knexHelper.showPrimaryKeys(this.state.connection, this.state.schema, this.state.entity);

        if (idField.length == 0) {
            this.throw(501, '不支持无主键的表' + this.state.entity);
        } else {

            let ids = this.params.id.split(';').map((e) => e.split('='));

            let params = {};
            params[idField[0]] = this.params.id;

            this.state.pkParams = this.params.id.search(';') == -1 ? params : util.parsePk(ids, idField);
            this.state.idField = idField;

            if (ids.length != idField.length) {
                this.throw(400, '主键错误,应为' + idField);
            }

            debug('Primary key is', this.state.idField, '=', this.state.pkParams);
            yield next;
        }

    })
    .get('/:id', function* () {

        let entity = yield this.state.connection(this.state.entity)
            .withSchema(this.state.schema)
            .where(this.state.pkParams).limit(1);

        if (entity.length == 0) {
            this.throw(404, '找不到' + JSON.stringify(this.state.pkParams) + '的数据');
        } else {
            this.body = entity[0];
        }

    })
    .get('/', function* () {

        let order = yield util.sortResolver.bind(this)();

        let limit = this.query.pageSize || this.query.count || 30;
        let page = this.query.page || 0;
        page = page < 0 ? 0 : page;

        let offset = page * limit;
        delete this.query['pageSize'];
        delete this.query['page'];
        delete this.query['_sort'];

        let joiner = (conn) => {
            return conn;
        };
        if (this.state.joinConditions) {
            joiner = (conn) => {
                return this.state.joinConditions.reduce((_conn, JoinCnd) => {
                    return _conn.joinRaw(JoinCnd)
                }, conn);
            }
        }

        let result = [];
        let count = 0;

        try {

            let sqlQuery = new httpQueryParser().parse(jsonFilter.ignore(this.query, ['_charset']));

            let query = this.state.connection
                .withSchema(this.state.schema)
                .select('*')
                .from(this.state.entity)
            if (order.length) {
                query = query.orderByRaw(order.join(','));
            }
            query = joiner(query).limit(limit)
                .offset(offset)
            result = yield util.appendClause(query, sqlQuery);

            count = yield util.appendClause(joiner(this.state.connection
                    .withSchema(this.state.schema)
                    .count('* as totalElements')
                    .from(this.state.entity)),
                sqlQuery);

            this.body = util.wrapPagingResult.call(this, result, {
                count: count[0].totalElements,
                page: page,
                limit: limit
            });

        } catch (e) {
            debug('Pagination error %o', e);
            this.throw(e.message, 400);
        }

    })
    .post('/', function* () {

        let future = co(function* () {
            let body = yield datetimeService.transformDatetimeFields(
                this.state.connection,
                this.state.schema,
                this.state.entity,
                this.request.body);

            let id = yield this.state.connection(this.state.entity)
                .withSchema(this.state.schema)
                .insert(body)
                .on('query', queryLogger.bind(this));

            let idField = yield this.knexHelper.showPrimaryKeys(this.state.connection, this.state.schema, this.state.entity);

            let pkParams = {};

            if (idField.length > 1) {
                idField.forEach((e) => {
                    if (this.request.body[e]) {
                        pkParams[e] = this.request.body[e];
                    }
                });
            } else if (idField.length == 1) {
                var pk = this.request.body[idField[0]];
                //Not an auto increment pk
                if (pk && (pk != id[0])) {
                    pkParams[idField[0]] = pk;
                } else {
                    pkParams[idField[0]] = id[0];
                }
            } else {
                //A strange table without pk...
                pkParams = this.request.body;
            }

            if(Object.keys(pkParams).length == 0) {
                pkParams = this.request.body;
            }

            debug('Pk is', pkParams);

            let entity = yield this.state.connection(this.state.entity)
                .withSchema(this.state.schema)
                .where(pkParams).limit(1);

            this.status = 201;
            this.body = entity[0];

        }.bind(this));

        //TODO queue
        if (this.state.async) {
            this.body = this.request.body;
            Promise.all([future]);
        } else {
            yield future;
        }

    })
    .post('/rows', importSupport(), function* () {

        let rows = typeof this.request.body === 'string' ? (JSON.parse(this.request.body)).rows : this.request.body.rows;

        let body = [];
        for (let i = 0, l = rows.length; i < l; i++) {
            let one = yield datetimeService.transformDatetimeFields(
                this.state.connection,
                this.state.schema,
                this.state.entity,
                rows[i]);
            body.push(one);
        }

        let id = yield this.state.connection(this.state.entity)
            .withSchema(this.state.schema)
            .insert(body)
            .on('query', queryLogger.bind(this));

        let auto = yield this.state.connection(this.state.entity)
            .select('*').from('information_schema.columns')
            .where({
                table_name: this.state.entity,
                table_schema: this.state.schema,
                COLUMN_KEY: "PRI",
                EXTRA: "auto_increment"
            });

        let entity = rows;

        if (auto.length) {
            let ids = [];
            for (let i = 0, l = rows.length; i < l; i++) {
                ids.push(id[0] + i);
            }

            debug('Pk is', auto);
            entity = yield this.state.connection(this.state.entity)
                .withSchema(this.state.schema)
                .whereIn(auto[0].COLUMN_NAME, ids);
        }

        this.status = 201;
        this.body = entity;
    })
    .put('/:id', updateFunction)
    .patch('/:id', updateFunction)
    .delete('/:id', function* () {
        try {
            let res = yield this.state.connection(this.state.entity)
                .withSchema(this.state.schema)
                .where(this.state.pkParams)
                .del();
            this.body = [{
                affects: res
            }];
            //.on('query', queryLogger.bind(this))
        } catch (e) {
            this.status = 500;
            this.body = e;
        }
    });

module.exports = router;