/**
 * Created by eggcanfly on 16/6/15.
 */
'use strict';

const debug = require('debug')('krh:route');
const route = require('koa-router');

const util = require('./util');
const SqlParser = require('../core/sql/sqlParser');
const knexHelper = require('../core/knex-helper');
const ObjStream = require('objstream')

let router = new route();
let sqlParser = new SqlParser();

router
    .param('schema', function* (schema, next) {
        if(this.state.schema) {
            return yield next;
        }

        debug('schema param', schema);
        //resolve path to state.schema for table usage
        if (!schema) {
            this.body = '缺少数据库schema';
            return this.status = 400;
        } else {
            this.state.schema = this.state.schema || schema;
            yield next;
        }
    })
;

router.get('/form', function* () {
    yield this.render('sqlExecutor');
});

let rejectNonQuery = function (sql) {
    if (sql.toLowerCase().search('select') == -1) {
        this.throw('只支持sql select', 400)
    }
};

let cleanParams = function (sql) {
    //log start
    let params = {};

    if (JSON.stringify(this.query) !== '{}') {
        params = this.query;
    } else if (JSON.stringify(this.request.body) !== '{}') {
        params = this.request.body;
    }

    try {
        this.state.mappingSql = sqlParser.cleanAndReplaceParams(sql, this.query, this.state.schema);
    } catch (err) {
        debug('Fail back to normal sql', `"${sql}"`);
        debug(err.message, `"${sql.substring(err.location.start.offset - 5, err.location.end.offset + 5)}"`);
        this.state.mappingSql = sql;
    }

    return this.state.mappingSql;
};

router
// 前置处理
    .all('*', function* (next) {

        if (!this.query['sql']) {
            this.query['sql'] = this.request.body.sql;
        }

        let accessTime = new Date();
        let source = this.headers['Datahub-Source'] || this.origin || 'unknown';
        let sql = this.query['sql'];

        //log start
        let logId = yield this.knexHelper.logSqlAccessing(
            this.ip,
            accessTime,
            -1,
            sql,
            JSON.stringify(this.query),
            source
        );

        yield next;

        let costTime = new Date().getTime() - accessTime.getTime();

        yield this.knexHelper.updateCostTime(logId, costTime);

        try {
            this.app.emit('sql-exec', [sql], [1, costTime])
        } catch (err) {
            debug('pp report error', err);
        }
    })
    .post('/', function* () {

        //支持事务 #18
        let sql = sqlParser.formatSql(this.query['sql'] || this.request.body['sql']);
        delete this.query['sql'];

        sql = cleanParams.bind(this)(sql);

        if (this.state.schema) {
            yield this.state.connection.raw('use :schema:', {schema: this.state.schema});
        }
        sql = sqlParser.cleanAndReplaceParams(this.state.mappingSql, this.query);
        this.body = this.state.connection
            .raw(sql, this.query).stream().pipe(new ObjStream());

    })
    .get('/', knexHelper.readOnlyTx.bind(this)(), function* () {

        let sql = sqlParser.formatSql(this.query['sql']);
        rejectNonQuery.bind(this)(sql);

        delete this.query['sql'];

        sql = cleanParams.bind(this)(sql);

        if (this.state.schema) {
            yield this.state.trx.raw('use :schema:', {schema: this.state.schema});
        }
        try {
            this.body = this.state.trx.raw(sql, this.query).stream().pipe(new ObjStream());
        } catch (err) {
            this.body = err;
        }

    })
    .get('/pagination(.*)', knexHelper.readOnlyTx.bind(this)(), function* () {

        let query = this.query;
        let sql = query.sql;
        rejectNonQuery.bind(this)(sql);
        delete this.query['sql'];

        sql = cleanParams.bind(this)(sql);

        if (this.state.schema) {
            yield this.state.trx.raw('use :schema:', {schema: this.state.schema});
        }

        try {
            let data = yield this.knexHelper.paginate(sqlParser.formatSql(sql),
                this.query,
                this.state.trx);
            this.body = util.wrapPagingResult(data.result, data.options);
        } catch (err) {
            this.body = err;
        }

    });

module.exports = router;