const HttpQueryParser = require('../core/http/http-query-parser');
const should = require('should');
const parser = new HttpQueryParser();
const qs = require('querystring');

describe("HttpQueryParser Simple query", function () {
    it("a=x should parsed to a = x", function () {
        parser.parse(qs.parse('a=x')).should.eql({
            a: {columnName: 'a', operatorAlias: '=', value: 'x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a=x&b=y should parsed to a = x and b = y ", function () {
        parser.parse(qs.parse('a=x&b=y')).should.eql({
            a: {columnName: 'a', operatorAlias: '=', value: 'x', condition: 'and', whereType: 'Where'},
            b: {columnName: 'b', operatorAlias: '=', value: 'y', condition: 'and', whereType: 'Where'}
        });
    });

    it("should be an empty query", function () {
        parser.parse(qs.parse('a=&b:in=')).should.eql({});
    });
});

describe("HttpQueryParser Query with operator", function () {
    it("a:contains=x should parsed to a like '%x%'", function () {
        parser.parse(qs.parse('a:contains=x')).should.eql({
            'a:contains': {columnName: 'a', operatorAlias: 'like', value: '%x%', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:like=x& should parsed to a like '%x%'", function () {
        parser.parse(qs.parse('a:like=x')).should.eql({
            'a:like': {columnName: 'a', operatorAlias: 'like', value: '%x%', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:startsWith=x& should parsed to a like 'x%'", function () {
        parser.parse(qs.parse('a:startsWith=x')).should.eql({
            'a:startsWith': {columnName: 'a', operatorAlias: 'like', value: 'x%', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:endsWith=x& should parsed to a like '%x'", function () {
        parser.parse(qs.parse('a:endsWith=x')).should.eql({
            'a:endsWith': {columnName: 'a', operatorAlias: 'like', value: '%x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:after=x& should parsed to a > x", function () {
        parser.parse(qs.parse('a:after=x')).should.eql({
            'a:after': {columnName: 'a', operatorAlias: '>', value: 'x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:gt=x& should parsed to a > x", function () {
        parser.parse(qs.parse('a:gt=x')).should.eql({
            'a:gt': {columnName: 'a', operatorAlias: '>', value: 'x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:gte=x& should parsed to a >= x", function () {
        parser.parse(qs.parse('a:gte=x')).should.eql({
            'a:gte': {columnName: 'a', operatorAlias: '>=', value: 'x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:before=x& should parsed to a < x", function () {
        parser.parse(qs.parse('a:before=x')).should.eql({
            'a:before': {columnName: 'a', operatorAlias: '<', value: 'x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:lt=x& should parsed to a < x", function () {
        parser.parse(qs.parse('a:lt=x')).should.eql({
            'a:lt': {columnName: 'a', operatorAlias: '<', value: 'x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:lte=x& should parsed to a <= x", function () {
        parser.parse(qs.parse('a:lte=x')).should.eql({
            'a:lte': {columnName: 'a', operatorAlias: '<=', value: 'x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:lte=x&b:contains=y&c=z should parsed to a <= x and b like '%y%' and c = z", function () {
        parser.parse(qs.parse('a:lte=x&b:contains=y&c=z')).should.eql({
            'a:lte': {columnName: 'a', operatorAlias: '<=', value: 'x', condition: 'and', whereType: 'Where'},
            'b:contains': {columnName: 'b', operatorAlias: 'like', value: '%y%', condition: 'and', whereType: 'Where'},
            'c': {columnName: 'c', operatorAlias: '=', value: 'z', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:lte=x&a:gte=y should parsed to a <= x and a >= y", function () {
        parser.parse(qs.parse('a:lte=x&a:gte=y')).should.eql({
            'a:lte': {columnName: 'a', operatorAlias: '<=', value: 'x', condition: 'and', whereType: 'Where'},
            'a:gte': {columnName: 'a', operatorAlias: '>=', value: 'y', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:in=1,2,xyz", function () {
        parser.parse(qs.parse('a:in=1,2,xyz')).should.eql({
            'a:in': {
                columnName: 'a',
                operatorAlias: 'in',
                value: ['1', '2', 'xyz'],
                condition: 'and',
                whereType: 'WhereIn'
            }
        });
    });

    it("a:notin=1,2,xyz", function () {
        parser.parse(qs.parse('a:notin=1,2,xyz')).should.eql({
            'a:notin': {
                columnName: 'a',
                operatorAlias: 'not in',
                value: ['1', '2', 'xyz'],
                condition: 'and',
                whereType: 'WhereNotIn'
            }
        });
    });

    it("a:in=1&a:in=2&a:in=xyz", function () {
        parser.parse(qs.parse('a:in=1&a:in=2&a:in=xyz')).should.eql({
            'a:in': {
                columnName: 'a',
                operatorAlias: 'in',
                value: ['1', '2', 'xyz'],
                condition: 'and',
                whereType: 'WhereIn'
            }
        });
    });

    it("a:between=1,100", function () {
        parser.parse(qs.parse('a:between=1,100')).should.eql({
            'a:between': {
                columnName: 'a',
                operatorAlias: 'between',
                value: [1, 100],
                condition: 'and',
                whereType: 'WhereBetween'
            }
        });
    });

    it("a:notbetween=40,50", function () {
        parser.parse(qs.parse('a:notbetween=40,50')).should.eql({
            'a:notbetween': {
                columnName: 'a',
                operatorAlias: 'not between',
                value: [40, 50],
                condition: 'and',
                whereType: 'WhereNotBetween'
            }
        });
    });
});

describe("HttpQueryParser Query with not", function () {

    it("a:not=x should parsed to a <> x", function () {
        parser.parse(qs.parse('a:not=x')).should.eql({
            'a:not': {columnName: 'a', operatorAlias: '<>', value: 'x', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:notStartsWith=x should parsed to a not like x%", function () {
        parser.parse(qs.parse('a:notStartsWith=x')).should.eql({
            'a:notStartsWith': {
                columnName: 'a',
                operatorAlias: 'not like',
                value: 'x%',
                condition: 'and',
                whereType: 'Where'
            }
        });
    });

    it("a:notEndsWith=x should parsed to a not like %x", function () {
        parser.parse(qs.parse('a:notEndsWith=x')).should.eql({
            'a:notEndsWith': {
                columnName: 'a',
                operatorAlias: 'not like',
                value: '%x',
                condition: 'and',
                whereType: 'Where'
            }
        });
    });

    it("a:notContains=x should parsed to a not like %x%", function () {
        parser.parse(qs.parse('a:notContains=x')).should.eql({
            'a:notContains': {
                columnName: 'a',
                operatorAlias: 'not like',
                value: '%x%',
                condition: 'and',
                whereType: 'Where'
            }
        });
    });

    it("a:notLike=x should parsed to a not like x%", function () {
        parser.parse(qs.parse('a:notLike=x')).should.eql({
            'a:notLike': {
                columnName: 'a',
                operatorAlias: 'not like',
                value: '%x%',
                condition: 'and',
                whereType: 'Where'
            }
        });
    });

});

describe("HttpQueryParser Query with or", function () {

    it("or:a=x should parsed to or a = x", function () {
        parser.parse(qs.parse('or:a=x')).should.eql({
            'or:a': {columnName: 'a', operatorAlias: '=', value: 'x', condition: 'or', whereType: 'Where'}
        });
    });

    it("or:a:contains=x should parsed to or a like '%x%'", function () {
        parser.parse(qs.parse('or:a:contains=x')).should.eql({
            'or:a:contains': {columnName: 'a', operatorAlias: 'like', value: '%x%', condition: 'or', whereType: 'Where'}
        });
    });

    it("or:a:contains=x&b:gt=y should parsed to or a like '%x%' and b > y", function () {
        parser.parse(qs.parse('or:a:contains=x&b:gt=y')).should.eql({
            'or:a:contains': {
                columnName: 'a',
                operatorAlias: 'like',
                value: '%x%',
                condition: 'or',
                whereType: 'Where'
            },
            'b:gt': {columnName: 'b', operatorAlias: '>', value: 'y', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:lte=x&or:a:gte=y should parsed to a <= x or a >= y", function () {
        parser.parse(qs.parse('a:lte=x&or:a:gte=y')).should.eql({
            'a:lte': {columnName: 'a', operatorAlias: '<=', value: 'x', condition: 'and', whereType: 'Where'},
            'or:a:gte': {columnName: 'a', operatorAlias: '>=', value: 'y', condition: 'or', whereType: 'Where'}
        });
    });

    it("a:lte=x&or:a:gte=y&b=z should parsed to a <= x or a >= y and b = z", function () {
        parser.parse(qs.parse('a:lte=x&or:a:gte=y&b=z')).should.eql({
            'a:lte': {columnName: 'a', operatorAlias: '<=', value: 'x', condition: 'and', whereType: 'Where'},
            'or:a:gte': {columnName: 'a', operatorAlias: '>=', value: 'y', condition: 'or', whereType: 'Where'},
            'b': {columnName: 'b', operatorAlias: '=', value: 'z', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:lte=x&or:a:gte=y&b=z&c:not=m should parsed to a <= x or a >= y and b = z and c <> m", function () {
        parser.parse(qs.parse('a:lte=x&or:a:gte=y&b=z&c:not=m')).should.eql({
            'a:lte': {columnName: 'a', operatorAlias: '<=', value: 'x', condition: 'and', whereType: 'Where'},
            'or:a:gte': {columnName: 'a', operatorAlias: '>=', value: 'y', condition: 'or', whereType: 'Where'},
            'b': {columnName: 'b', operatorAlias: '=', value: 'z', condition: 'and', whereType: 'Where'},
            'c:not': {columnName: 'c', operatorAlias: '<>', value: 'm', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:isnotempty&b:isnull should parsed to z is not null or a >= y and b = z and c <> m", function () {
        parser.parse(qs.parse('a:isnotempty&b:isnull')).should.eql({
            'a:isnotempty': {columnName: 'a', operatorAlias: '<>', value: '', condition: 'and', whereType: 'Where'},
            'b:isnull': {
                columnName: 'b',
                operatorAlias: 'is null',
                value: HttpQueryParser.IS_NULL_VALUE,
                condition: 'and',
                whereType: 'WhereNull'
            }
        });
    });

});

describe("HttpQueryParser Query with null", function () {

    it("z:isnull=&or:a:gte=y&b=z&c:not=m should parsed to a is null", function () {
        parser.parse(qs.parse('z:isnull=&or:a:gte=y&b=z&c:not=m')).should.eql({
            'z:isnull': {
                columnName: 'z',
                operatorAlias: 'is null',
                value: HttpQueryParser.IS_NULL_VALUE,
                condition: 'and',
                whereType: 'WhereNull'
            },
            'or:a:gte': {columnName: 'a', operatorAlias: '>=', value: 'y', condition: 'or', whereType: 'Where'},
            'b': {columnName: 'b', operatorAlias: '=', value: 'z', condition: 'and', whereType: 'Where'},
            'c:not': {columnName: 'c', operatorAlias: '<>', value: 'm', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:isnotnull= should parsed to a is not null", function () {
        parser.parse(qs.parse('a:isnotnull=')).should.eql({
            'a:isnotnull': {
                columnName: 'a',
                operatorAlias: 'is not null',
                value: HttpQueryParser.IS_NOT_NULL_VALUE,
                condition: 'and',
                whereType: 'WhereNotNull'
            }
        });
    });

    it("a:isempty= should parsed to a = ''", function () {
        parser.parse(qs.parse('a:isempty=')).should.eql({
            'a:isempty': {columnName: 'a', operatorAlias: '=', value: '', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:isnotempty= should parsed to a <> ''", function () {
        parser.parse(qs.parse('a:isnotempty=')).should.eql({
            'a:isnotempty': {columnName: 'a', operatorAlias: '<>', value: '', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:isnotempty should parsed to a <> ''", function () {
        parser.parse(qs.parse('a:isnotempty')).should.eql({
            'a:isnotempty': {columnName: 'a', operatorAlias: '<>', value: '', condition: 'and', whereType: 'Where'}
        });
    });

    it("a:isnotempty&b:isnull should parsed to a <> '' and b is null", function () {
        parser.parse(qs.parse('a:isnotempty&b:isnull')).should.eql({
            'a:isnotempty': {columnName: 'a', operatorAlias: '<>', value: '', condition: 'and', whereType: 'Where'},
            'b:isnull': {
                columnName: 'b',
                operatorAlias: 'is null',
                value: HttpQueryParser.IS_NULL_VALUE,
                condition: 'and',
                whereType: 'WhereNull'
            }
        });
    });
})

describe("realword test", function () {
    it("long in query", function () {
        parser.parse(qs.parse("f_status:lt=1000&page=0&pageSize=10&f_inner_id:in=1119512423703904256,1119512423783596032,1119512423808761856")).should.eql({
            'f_status:lt': {
                columnName: 'f_status',
                operatorAlias: '<',
                value: '1000',
                condition: 'and',
                whereType: 'Where'
            },
            'page': {columnName: 'page', operatorAlias: '=', value: '0', condition: 'and', whereType: 'Where'},
            'pageSize': {columnName: 'pageSize', operatorAlias: '=', value: '10', condition: 'and', whereType: 'Where'},
            'f_inner_id:in': {
                columnName: 'f_inner_id',
                operatorAlias: 'in',
                value: ['1119512423703904256', '1119512423783596032', '1119512423808761856'],
                condition: 'and',
                whereType: 'WhereIn'
            },
        })
    });
});